@@ -113,14 +113,16 @@ inline const long getBeagleCPUFlags<double>(){ return BEAGLE_FLAG_COMPUTATION_SY
                                                       BEAGLE_FLAG_THREADING_NONE |
                                                       BEAGLE_FLAG_PROCESSOR_CPU |
                                                       BEAGLE_FLAG_PRECISION_DOUBLE |
-                                                      BEAGLE_FLAG_VECTOR_NONE; };
+                                                      BEAGLE_FLAG_VECTOR_NONE |
+                                                      BEAGLE_FLAG_FRAMEWORK_CPU; };
 
 template<>
 inline const long getBeagleCPUFlags<float>(){ return BEAGLE_FLAG_COMPUTATION_SYNCH |
                                                      BEAGLE_FLAG_THREADING_NONE |
                                                      BEAGLE_FLAG_PROCESSOR_CPU |
                                                      BEAGLE_FLAG_PRECISION_SINGLE |
-                                                     BEAGLE_FLAG_VECTOR_NONE; };
+                                                     BEAGLE_FLAG_VECTOR_NONE |
+                                                     BEAGLE_FLAG_FRAMEWORK_CPU; };
 
 
 
@@ -200,6 +202,7 @@ int BeagleCPUImpl<BEAGLE_CPU_GENERIC>::createInstance(int tipCount,
                                   int categoryCount,
                                   int scaleBufferCount,
                                   int resourceNumber,
+                                  int pluginResourceNumber,
                                   long preferenceFlags,
                                   long requirementFlags) {
     if (DEBUGGING_OUTPUT)
@@ -1079,7 +1082,7 @@ int BeagleCPUImpl<BEAGLE_CPU_GENERIC>::calcRootLogLikelihoodsMulti(const int* bu
         *outSumLogLikelihood += outLogLikelihoodsTmp[i] * gPatternWeights[i];
     }
 
-    if (!(*outSumLogLikelihood - *outSumLogLikelihood == 0.0))
+    if (*outSumLogLikelihood != *outSumLogLikelihood)
         returnCode = BEAGLE_ERROR_FLOATING_POINT;
     
     return returnCode;
@@ -1142,9 +1145,8 @@ int BeagleCPUImpl<BEAGLE_CPU_GENERIC>::calcRootLogLikelihoods(const int bufferIn
         *outSumLogLikelihood += outLogLikelihoodsTmp[i] * gPatternWeights[i];
     }
 
-    if (!(*outSumLogLikelihood - *outSumLogLikelihood == 0.0))
+    if (*outSumLogLikelihood != *outSumLogLikelihood)
         returnCode = BEAGLE_ERROR_FLOATING_POINT;
-
     
     // TODO: merge the three kPatternCount loops above into one
 
@@ -1291,11 +1293,9 @@ BEAGLE_CPU_TEMPLATE
                                           outSumLogLikelihood);
 		} else {
             fprintf(stderr,"BeagleCPUImpl::calculateEdgeLogLikelihoods not yet implemented for count > 1 and derivatives\n");
-        }
-            
-
+        }            
     }
-
+    return BEAGLE_SUCCESS;
 }
 
 BEAGLE_CPU_TEMPLATE
@@ -1408,7 +1408,7 @@ int BeagleCPUImpl<BEAGLE_CPU_GENERIC>::calcEdgeLogLikelihoods(const int parIndex
         *outSumLogLikelihood += outLogLikelihoodsTmp[i] * gPatternWeights[i];
     }
 
-    if (!(*outSumLogLikelihood - *outSumLogLikelihood == 0.0))
+    if (*outSumLogLikelihood != *outSumLogLikelihood)
         returnCode = BEAGLE_ERROR_FLOATING_POINT;
     
     return returnCode;
@@ -1561,7 +1561,7 @@ int BeagleCPUImpl<BEAGLE_CPU_GENERIC>::calcEdgeLogLikelihoodsMulti(const int* pa
         *outSumLogLikelihood += outLogLikelihoodsTmp[i] * gPatternWeights[i];
     }
     
-    if (!(*outSumLogLikelihood - *outSumLogLikelihood == 0.0))
+    if (*outSumLogLikelihood != *outSumLogLikelihood)
         returnCode = BEAGLE_ERROR_FLOATING_POINT;
     
     return returnCode;
@@ -1677,7 +1677,7 @@ int BeagleCPUImpl<BEAGLE_CPU_GENERIC>::calcEdgeLogLikelihoodsFirstDeriv(const in
         *outSumFirstDerivative += outFirstDerivativesTmp[i] * gPatternWeights[i];
     }
     
-    if (!(*outSumLogLikelihood - *outSumLogLikelihood == 0.0))
+    if (*outSumLogLikelihood != *outSumLogLikelihood)
         returnCode = BEAGLE_ERROR_FLOATING_POINT;
 
     return returnCode;
@@ -1806,7 +1806,7 @@ int BeagleCPUImpl<BEAGLE_CPU_GENERIC>::calcEdgeLogLikelihoodsSecondDeriv(const i
         *outSumSecondDerivative += outSecondDerivativesTmp[i] * gPatternWeights[i];
     }
 
-    if (!(*outSumLogLikelihood - *outSumLogLikelihood == 0.0))
+    if (*outSumLogLikelihood != *outSumLogLikelihood)
         returnCode = BEAGLE_ERROR_FLOATING_POINT;
     
     return returnCode;
@@ -2252,7 +2252,7 @@ void* BeagleCPUImpl<BEAGLE_CPU_GENERIC>::mallocAligned(size_t size) {
     #if (T_PAD == 1)	
         const size_t align = 32;
     #else // T_PAD == 2
-        const size_t align = 16;
+        const size_t align = 32; // Changed from 16 (under SSE) to ensure AVX alignment
     #endif
 	int res;
 	res = posix_memalign(&ptr, align, size);
@@ -2277,6 +2277,7 @@ BeagleImpl* BeagleCPUImplFactory<BEAGLE_CPU_FACTORY_GENERIC>::createImpl(int tip
                                              int categoryCount,
                                              int scaleBufferCount,
                                              int resourceNumber,
+                                             int pluginResourceNumber,
                                              long preferenceFlags,
                                              long requirementFlags,
                                              int* errorCode) {
@@ -2287,7 +2288,9 @@ BeagleImpl* BeagleCPUImplFactory<BEAGLE_CPU_FACTORY_GENERIC>::createImpl(int tip
         *errorCode =
             impl->createInstance(tipCount, partialsBufferCount, compactBufferCount, stateCount,
                                  patternCount, eigenBufferCount, matrixBufferCount,
-                                 categoryCount,scaleBufferCount, resourceNumber, preferenceFlags, requirementFlags);
+                                 categoryCount,scaleBufferCount, resourceNumber,
+                                 pluginResourceNumber, 
+                                 preferenceFlags, requirementFlags);
         if (*errorCode == BEAGLE_SUCCESS) {
             return impl;
         }
@@ -2321,7 +2324,8 @@ const long BeagleCPUImplFactory<BEAGLE_CPU_FACTORY_GENERIC>::getFlags() {
                  BEAGLE_FLAG_VECTOR_NONE |
                  BEAGLE_FLAG_SCALERS_LOG | BEAGLE_FLAG_SCALERS_RAW |
                  BEAGLE_FLAG_EIGEN_COMPLEX | BEAGLE_FLAG_EIGEN_REAL |
-                 BEAGLE_FLAG_INVEVEC_STANDARD | BEAGLE_FLAG_INVEVEC_TRANSPOSED;
+                 BEAGLE_FLAG_INVEVEC_STANDARD | BEAGLE_FLAG_INVEVEC_TRANSPOSED |
+                 BEAGLE_FLAG_FRAMEWORK_CPU;
 	if (DOUBLE_PRECISION)
 		flags |= BEAGLE_FLAG_PRECISION_DOUBLE;
 	else
