@@ -38,7 +38,6 @@
 
 #include "libhmsbeagle/beagle.h"
 #include "libhmsbeagle/GPU/GPUImplDefs.h"
-#include "libhmsbeagle/GPU/BeagleGPUImpl.h"
 #include "libhmsbeagle/GPU/GPUImplHelper.h"
 #include "libhmsbeagle/GPU/KernelLauncher.h"
 #include "libhmsbeagle/GPU/GPUInterface.h"
@@ -47,6 +46,12 @@
 namespace beagle {
 namespace gpu {
 
+#ifdef CUDA
+	namespace cuda {
+#else
+	namespace opencl {
+#endif
+
 BEAGLE_GPU_TEMPLATE
 BeagleGPUImpl<BEAGLE_GPU_GENERIC>::BeagleGPUImpl() {
     
@@ -124,6 +129,7 @@ BeagleGPUImpl<BEAGLE_GPU_GENERIC>::~BeagleGPUImpl() {
             gpu->FreeMemory(dFrequencies[i]);
         }
 
+        // TODO: free subpointers
         gpu->FreeMemory(dMatrices[0]);
         
         if (kFlags & BEAGLE_FLAG_SCALING_DYNAMIC) {
@@ -227,7 +233,8 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::createInstance(int tipCount,
                                   int matrixCount,
                                   int categoryCount,
                                   int scaleBufferCount,
-                                  int iResourceNumber,
+                                  int globalResourceNumber,
+                                  int pluginResourceNumber,
                                   long preferenceFlags,
                                   long requirementFlags) {
     
@@ -247,46 +254,123 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::createInstance(int tipCount,
     kCategoryCount = categoryCount;
     kScaleBufferCount = scaleBufferCount;
     
-    resourceNumber = iResourceNumber;
+    resourceNumber = globalResourceNumber;
 
     kTipPartialsBufferCount = kTipCount - kCompactBufferCount;
     kBufferCount = kPartialsBufferCount + kCompactBufferCount;
 
     kInternalPartialsBufferCount = kBufferCount - kTipCount;
     
-    if (kStateCount <= 4)
+    if (kStateCount <= 4) {
         kPaddedStateCount = 4;
-    else if (kStateCount <= 16)
+    } else if (kStateCount <= 16) {
         kPaddedStateCount = 16;
-    else if (kStateCount <= 32)
+    } else if (kStateCount <= 32) {
         kPaddedStateCount = 32;
-    else if (kStateCount <= 48)
-    	kPaddedStateCount = 48;    
-    else if (kStateCount <= 64)
+    } else if (kStateCount <= 48) {
+    	kPaddedStateCount = 48;  
+    } else if (kStateCount <= 64) {
         kPaddedStateCount = 64;
-    else if (kStateCount <= 80)
+    } else if (kStateCount <= 80) {
 		kPaddedStateCount = 80;
-    else if (kStateCount <= 128)
+    } else if (kStateCount <= 128) {
         kPaddedStateCount = 128;
-    else if (kStateCount <= 192)
+    } else if (kStateCount <= 192){ 
         kPaddedStateCount = 192;
-    else
+    } else {
         kPaddedStateCount = kStateCount + kStateCount % 16;
-        
-    // Make sure that kPaddedPatternCount + paddedPatterns is multiple of 4 for DNA model
-    int paddedPatterns = 0;
-    if (kPaddedStateCount == 4 && kPatternCount % 4 != 0) 
-        paddedPatterns = 4 - kPatternCount % 4;
-            
-    // TODO Should do something similar for 4 < kStateCount <= 8 as well
+    }
+
+    gpu = new GPUInterface();
     
+    gpu->Initialize();
+
+    int numDevices = 0;
+    numDevices = gpu->GetDeviceCount();
+    if (numDevices == 0) {
+        fprintf(stderr, "Error: No GPU devices\n");
+        return BEAGLE_ERROR_NO_RESOURCE;
+    }
+    if (pluginResourceNumber > numDevices) {
+        fprintf(stderr,"Error: Trying to initialize device # %d (which does not exist)\n",resourceNumber);
+        return BEAGLE_ERROR_NO_RESOURCE;
+    }
+
+    int paddedPatterns = 0;
+    // Make sure that kPaddedPatternCount + paddedPatterns is multiple of 4 for DNA model
+    if (kPaddedStateCount == 4 && kPatternCount % 4 != 0) 
+        paddedPatterns = 4 - kPatternCount % 4;    
+    // TODO Should do something similar for 4 < kStateCount <= 8 as well
+
+    bool CPUImpl = false;
+
+#ifdef FW_OPENCL
+    BeagleDeviceImplementationCodes deviceCode = gpu->GetDeviceImplementationCode(pluginResourceNumber);
+    
+    // TODO: Apple OpenCL on CPU for state count > 128
+    if (deviceCode == BEAGLE_OPENCL_DEVICE_APPLE_CPU && kPaddedStateCount > 128) {
+        return BEAGLE_ERROR_NO_IMPLEMENTATION;
+    }
+
+    // TODO: AMD GPU implementation for high state and category counts
+    if ((deviceCode == BEAGLE_OPENCL_DEVICE_APPLE_AMD_GPU || 
+        deviceCode == BEAGLE_OPENCL_DEVICE_AMD_GPU) &&
+        ((kPaddedStateCount > 64 && kCategoryCount > 2) || 
+          (kPaddedStateCount == 192 && kCategoryCount > 1))) {
+        return BEAGLE_ERROR_NO_IMPLEMENTATION;
+    }
+
+    if (deviceCode == BEAGLE_OPENCL_DEVICE_INTEL_CPU ||
+        deviceCode == BEAGLE_OPENCL_DEVICE_INTEL_MIC ||
+        deviceCode == BEAGLE_OPENCL_DEVICE_AMD_CPU ||
+        deviceCode == BEAGLE_OPENCL_DEVICE_APPLE_CPU) {
+        
+        CPUImpl = true;
+
+        int patternBlockSize = 0;
+        int id = (kFlags & BEAGLE_FLAG_PRECISION_DOUBLE ?
+                  kPaddedStateCount : (-1 * kPaddedStateCount));
+    
+        switch(id) {
+            case   -4: patternBlockSize = PATTERN_BLOCK_SIZE_DP_4_CPU; break;
+            case  -16: patternBlockSize = PATTERN_BLOCK_SIZE_DP_16;    break;
+            case  -32: patternBlockSize = PATTERN_BLOCK_SIZE_DP_32;    break;
+            case  -48: patternBlockSize = PATTERN_BLOCK_SIZE_DP_48;    break;
+            case  -64: patternBlockSize = PATTERN_BLOCK_SIZE_DP_64;    break;
+            case  -80: patternBlockSize = PATTERN_BLOCK_SIZE_DP_80;    break;
+            case -128: patternBlockSize = PATTERN_BLOCK_SIZE_DP_128;   break;
+            case -192: patternBlockSize = PATTERN_BLOCK_SIZE_DP_192;   break;
+            case    4: patternBlockSize = PATTERN_BLOCK_SIZE_SP_4_CPU; break;
+            case   16: patternBlockSize = PATTERN_BLOCK_SIZE_SP_16;    break;
+            case   32: patternBlockSize = PATTERN_BLOCK_SIZE_SP_32;    break;
+            case   48: patternBlockSize = PATTERN_BLOCK_SIZE_SP_48;    break;
+            case   64: patternBlockSize = PATTERN_BLOCK_SIZE_SP_64;    break;
+            case   80: patternBlockSize = PATTERN_BLOCK_SIZE_SP_80;    break;
+            case  128: patternBlockSize = PATTERN_BLOCK_SIZE_SP_128;   break;
+            case  192: patternBlockSize = PATTERN_BLOCK_SIZE_SP_192;   break;
+        }
+    
+        // pad patterns for CPU/MIC implementation
+        if (patternBlockSize != 0 && kPatternCount % patternBlockSize) {
+            paddedPatterns = patternBlockSize - (kPatternCount % patternBlockSize);
+        }
+    }
+#endif
+
     kPaddedPatternCount = kPatternCount + paddedPatterns;
     
     int resultPaddedPatterns = 0;
-    int patternBlockSizeFour = (kFlags & BEAGLE_FLAG_PRECISION_DOUBLE ? PATTERN_BLOCK_SIZE_DP_4 : PATTERN_BLOCK_SIZE_SP_4);
-    if (kPaddedStateCount == 4 && kPaddedPatternCount % patternBlockSizeFour != 0)
-        resultPaddedPatterns = patternBlockSizeFour - kPaddedPatternCount % patternBlockSizeFour;
-        
+
+    if (!CPUImpl) {
+        int patternBlockSizeFour = (kFlags & BEAGLE_FLAG_PRECISION_DOUBLE ? PATTERN_BLOCK_SIZE_DP_4 : PATTERN_BLOCK_SIZE_SP_4);
+        if (kPaddedStateCount == 4 && kPaddedPatternCount % patternBlockSizeFour != 0)
+            resultPaddedPatterns = patternBlockSizeFour - kPaddedPatternCount % patternBlockSizeFour;
+    }
+    
+#ifdef BEAGLE_DEBUG_VALUES
+    printf("kPatternCount %d, paddedPatterns %d, resultPaddedPatterns %d, kPaddedPatternCount %d\n", kPatternCount, paddedPatterns, resultPaddedPatterns, kPaddedPatternCount);
+#endif
+
     kScaleBufferSize = kPaddedPatternCount;
     
     kFlags = 0;
@@ -340,25 +424,14 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::createInstance(int tipCount,
         
     kLastCompactBufferIndex = -1;
     kLastTipPartialsBufferIndex = -1;
-        
-    gpu = new GPUInterface();
-    
-    gpu->Initialize();
-    
-    int numDevices = 0;
-    numDevices = gpu->GetDeviceCount();
-    if (numDevices == 0) {
-        fprintf(stderr, "Error: No GPU devices\n");
-        return BEAGLE_ERROR_NO_RESOURCE;
-    }
-    if (resourceNumber > numDevices) {
-        fprintf(stderr,"Error: Trying to initialize device # %d (which does not exist)\n",resourceNumber);
-        return BEAGLE_ERROR_NO_RESOURCE;
-    }
     
     // TODO: recompiling kernels for every instance, probably not ideal
-    gpu->SetDevice(resourceNumber-1,kPaddedStateCount,kCategoryCount,kPaddedPatternCount,kFlags);
-      
+    gpu->SetDevice(pluginResourceNumber,kPaddedStateCount,kCategoryCount,kPaddedPatternCount, kPatternCount,kFlags);
+
+#ifdef FW_OPENCL
+    kFlags |= gpu->GetDeviceTypeFlag(pluginResourceNumber);
+#endif
+
     int ptrQueueLength = kMatrixCount * kCategoryCount * 3;
     if (kPartialsBufferCount > ptrQueueLength)
         ptrQueueLength = kPartialsBufferCount;
@@ -382,17 +455,17 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::createInstance(int tipCount,
     sizeof(int) * kCompactBufferCount * kPaddedPatternCount + // dCompactBuffers
     sizeof(GPUPtr) * ptrQueueLength;  // dPtrQueue
     
-    
-    unsigned int availableMem = gpu->GetAvailableMemory();
-    
+#ifdef CUDA
+        unsigned int availableMem = gpu->GetAvailableMemory();
 #ifdef BEAGLE_DEBUG_VALUES
-    fprintf(stderr, "     needed memory: %d\n", neededMemory);
-    fprintf(stderr, "  available memory: %d\n", availableMem);
+        fprintf(stderr, "     needed memory: %d\n", neededMemory);
+        fprintf(stderr, "  available memory: %d\n", availableMem);
 #endif     
-    
-    if (availableMem < neededMemory) 
-        return BEAGLE_ERROR_OUT_OF_MEMORY;
-    
+        // TODO: fix memory check on CUDA and implement for OpenCL
+        // if (availableMem < neededMemory) 
+        //     return BEAGLE_ERROR_OUT_OF_MEMORY;
+#endif
+
     kernels = new KernelLauncher(gpu);
     
     // TODO: only allocate if necessary on the fly
@@ -415,33 +488,40 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::createInstance(int tipCount,
     dFrequencies = (GPUPtr*) calloc(sizeof(GPUPtr),kEigenDecompCount);
     
     dMatrices = (GPUPtr*) malloc(sizeof(GPUPtr) * kMatrixCount);
-    dMatrices[0] = gpu->AllocateMemory(kMatrixCount * kMatrixSize * kCategoryCount * sizeof(Real));
-    
-#ifdef CUDA
+
+    size_t ptrIncrement = gpu->AlignMemOffset(kMatrixSize * kCategoryCount * sizeof(Real));
+    dMatrices[0] = gpu->AllocateMemory(kMatrixCount * ptrIncrement);
     for (int i = 1; i < kMatrixCount; i++) {
-        dMatrices[i] = dMatrices[i-1] + kMatrixSize * kCategoryCount * sizeof(Real);
+        dMatrices[i] = gpu->CreateSubPointer(dMatrices[0], ptrIncrement*i, ptrIncrement);
     }
     
     if (kScaleBufferCount > 0) {
         if (kFlags & BEAGLE_FLAG_SCALING_AUTO) {        
             dScalingFactors = (GPUPtr*) malloc(sizeof(GPUPtr) * kScaleBufferCount);
-            dScalingFactors[0] =  gpu->AllocateMemory(sizeof(signed char) * kScaleBufferSize * kScaleBufferCount); // TODO: char won't work for double-precision
+            ptrIncrement = gpu->AlignMemOffset(kScaleBufferSize * sizeof(signed char)); // TODO: char won't work for double-precision
+            dScalingFactors[0] =  gpu->AllocateMemory(ptrIncrement * kScaleBufferCount);
             for (int i=1; i < kScaleBufferCount; i++)
-                dScalingFactors[i] = dScalingFactors[i-1] + kScaleBufferSize * sizeof(signed char);
+                dScalingFactors[i] = gpu->CreateSubPointer(dScalingFactors[0], ptrIncrement*i, ptrIncrement);
         } else if (kFlags & BEAGLE_FLAG_SCALING_DYNAMIC) {
+#ifdef CUDA
             dScalingFactors = (GPUPtr*) calloc(sizeof(GPUPtr), kScaleBufferCount);
             dScalingFactorsMaster = (GPUPtr*) calloc(sizeof(GPUPtr), kScaleBufferCount);
             hRescalingTrigger = (int*) gpu->AllocatePinnedHostMemory(sizeof(int), false, true);
-            dRescalingTrigger = gpu->GetDevicePointer((void*) hRescalingTrigger);
+            dRescalingTrigger = gpu->GetDeviceHostPointer((void*) hRescalingTrigger);
+#else
+            return BEAGLE_ERROR_NO_IMPLEMENTATION;
+#endif
         } else {
             dScalingFactors = (GPUPtr*) malloc(sizeof(GPUPtr) * kScaleBufferCount);
-            dScalingFactors[0] = gpu->AllocateMemory(kScaleBufferSize * kScaleBufferCount * sizeof(Real));
+            ptrIncrement = gpu->AlignMemOffset(kScaleBufferSize * sizeof(Real));
+            kScaleBufferSize = ptrIncrement / sizeof(Real);
+            dScalingFactors[0] = gpu->AllocateMemory(ptrIncrement * kScaleBufferCount);
             for (int i=1; i < kScaleBufferCount; i++) {
-                dScalingFactors[i] = dScalingFactors[i-1] + kScaleBufferSize * sizeof(Real);
+                dScalingFactors[i] = gpu->CreateSubPointer(dScalingFactors[0], ptrIncrement*i, ptrIncrement);
             }
-        }        
+        }
     }
-#endif
+
     
     for(int i=0; i<kEigenDecompCount; i++) {
     	dEvec[i] = gpu->AllocateMemory(kMatrixSize * sizeof(Real));
@@ -506,8 +586,8 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::createInstance(int tipCount,
 	hPatternWeightsCache = (Real*) gpu->MallocHost(sizeof(double) * kPatternCount);
     checkHostMemory(hPatternWeightsCache);
     
-	dMaxScalingFactors = gpu->AllocateMemory(kPaddedPatternCount * sizeof(Real));
-	dIndexMaxScalingFactors = gpu->AllocateMemory(kPaddedPatternCount * sizeof(int));
+	dMaxScalingFactors = gpu->AllocateMemory((kPaddedPatternCount + resultPaddedPatterns) * sizeof(Real));
+	dIndexMaxScalingFactors = gpu->AllocateMemory((kPaddedPatternCount + resultPaddedPatterns) * sizeof(unsigned int));
     
     if (kFlags & BEAGLE_FLAG_SCALING_AUTO) {
         dAccumulatedScalingFactors = gpu->AllocateMemory(sizeof(int) * kScaleBufferSize);
@@ -518,17 +598,20 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::createInstance(int tipCount,
 #endif
     
     kInitialized = 1;
-    
+
+#ifdef CUDA
 #ifdef BEAGLE_DEBUG_VALUES
     gpu->Synchronize();
     int usedMemory = availableMem - gpu->GetAvailableMemory();
     fprintf(stderr, "actual used memory: %d\n", usedMemory);
     fprintf(stderr, "        difference: %d\n\n", usedMemory-neededMemory);
 #endif
-    
+#endif
+
     return BEAGLE_SUCCESS;
 }
 
+#ifdef CUDA
 template<>
 char* BeagleGPUImpl<double>::getInstanceName() {
 	return (char*) "CUDA-Double";
@@ -538,6 +621,17 @@ template<>
 char* BeagleGPUImpl<float>::getInstanceName() {
 	return (char*) "CUDA-Single";
 }
+#elif defined(FW_OPENCL)
+template<>
+char* BeagleGPUImpl<double>::getInstanceName() {
+    return (char*) "OpenCL-Double";
+}
+
+template<>
+char* BeagleGPUImpl<float>::getInstanceName() {
+    return (char*) "OpenCL-Single";
+}
+#endif
 
 BEAGLE_GPU_TEMPLATE
 int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::getInstanceDetails(BeagleInstanceDetails* returnInfo) {
@@ -545,11 +639,17 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::getInstanceDetails(BeagleInstanceDetails*
         returnInfo->resourceNumber = resourceNumber;
         returnInfo->flags = BEAGLE_FLAG_COMPUTATION_SYNCH |
                             BEAGLE_FLAG_THREADING_NONE |
-                            BEAGLE_FLAG_VECTOR_NONE |
-                            BEAGLE_FLAG_PROCESSOR_GPU;
+                            BEAGLE_FLAG_VECTOR_NONE;
         Real r = 0;
         modifyFlagsForPrecision(&(returnInfo->flags), r);
 
+#ifdef CUDA
+        kFlags |= BEAGLE_FLAG_FRAMEWORK_CUDA;
+        kFlags |= BEAGLE_FLAG_PROCESSOR_GPU;
+#elif defined(FW_OPENCL)
+        kFlags |= BEAGLE_FLAG_FRAMEWORK_OPENCL;
+#endif
+
         returnInfo->flags |= kFlags;        
         
         returnInfo->implName = getInstanceName();
@@ -1068,8 +1168,6 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::convolveTransitionMatrices(const int* fir
 		int indexOffset = kMatrixSize * kCategoryCount;
 		int categoryOffset = kMatrixSize;
 
-#ifdef CUDA
-
 		for (int i = 0; i < matrixCount; i++) {
 			for (int j = 0; j < kCategoryCount; j++) {
 
@@ -1086,7 +1184,6 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::convolveTransitionMatrices(const int* fir
 
 		kernels->ConvolveTransitionMatrices(dMatrices[0], dPtrQueue, totalMatrixCount);
 
-#endif// END: if CUDA
 	}//END: count check
 
 #ifdef BEAGLE_DEBUG_FLOW
@@ -1111,11 +1208,9 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::updateTransitionMatrices(int eigenIndex,
         // TODO: improve performance of calculation of derivatives
         int totalCount = 0;
         
-        int indexOffset =  kMatrixSize * kCategoryCount;
+        int indexOffset =  gpu->AlignMemOffset(kMatrixSize * kCategoryCount * sizeof(Real)) / sizeof(Real);
         int categoryOffset = kMatrixSize;
         
-    #ifdef CUDA
-        
         if (firstDerivativeIndices == NULL && secondDerivativeIndices == NULL) {
             for (int i = 0; i < count; i++) {        
                 for (int j = 0; j < kCategoryCount; j++) {
@@ -1172,31 +1267,9 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::updateTransitionMatrices(int eigenIndex,
                                                       dEigenValues[eigenIndex], dDistanceQueue, totalCount);        
         }
         
-        
-    #else
-        // TODO: update OpenCL implementation with derivs
-        
-        for (int i = 0; i < count; i++) {        
-            for (int j = 0; j < kCategoryCount; j++) {
-                hDistanceQueue[totalCount] = (Real) (edgeLengths[i] * hCategoryRates[j]);
-                totalCount++;
-            }
-        }
-        
-        gpu->MemcpyHostToDevice(dDistanceQueue, hDistanceQueue, sizeof(Real) * totalCount);
-        
-        // Set-up and call GPU kernel
-        for (int i = 0; i < count; i++) {        
-            kernels->GetTransitionProbabilitiesSquare(dMatrices[probabilityIndices[i]], dEvec[eigenIndex], dIevc[eigenIndex], 
-                                                      dEigenValues[eigenIndex], dDistanceQueue, kCategoryCount,
-                                                      i * kCategoryCount);
-        }
-            
-    #endif
-        
     #ifdef BEAGLE_DEBUG_VALUES
         Real r = 0;
-        for (int i = 0; i < 1; i++) {
+        for (int i = 0; i < count; i++) {
             fprintf(stderr, "dMatrices[probabilityIndices[%d]]  (hDQ = %1.5e, eL = %1.5e) =\n", i,hDistanceQueue[i], edgeLengths[i]);        
             gpu->PrintfDeviceVector(dMatrices[probabilityIndices[i]], kMatrixSize * kCategoryCount, r);
             for(int j=0; j<kCategoryCount; j++)
@@ -1417,15 +1490,8 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::accumulateScaleFactors(const int* scaling
         gpu->MemcpyHostToDevice(dPtrQueue, hPtrQueue, sizeof(unsigned int) * count);
         
 
-    #ifdef CUDA    
-        // Compute scaling factors at the root
-        kernels->AccumulateFactorsDynamicScaling(dScalingFactors[0], dPtrQueue, dScalingFactors[cumulativeScalingIndex], count, kPaddedPatternCount);
-    #else // OpenCL
-//        for (int i = 0; i < count; i++) {
-//            kernels->AccumulateFactorsDynamicScaling(dScalingFactors[scalingIndices[i]], dScalingFactors[cumulativeScalingIndex],
-//                                                     1, kPaddedPatternCount);
-//        }
-    #endif
+    // Compute scaling factors at the root
+    kernels->AccumulateFactorsDynamicScaling(dScalingFactors[0], dPtrQueue, dScalingFactors[cumulativeScalingIndex], count, kPaddedPatternCount);
     }
     
 #ifdef BEAGLE_DEBUG_SYNCH    
@@ -1466,16 +1532,9 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::removeScaleFactors(const int* scalingIndi
         hPtrQueue[n] = scalingIndices[n] * kScaleBufferSize;
     gpu->MemcpyHostToDevice(dPtrQueue, hPtrQueue, sizeof(unsigned int) * count);
     
-#ifdef CUDA    
     // Compute scaling factors at the root
     kernels->RemoveFactorsDynamicScaling(dScalingFactors[0], dPtrQueue, dScalingFactors[cumulativeScalingIndex],
                                          count, kPaddedPatternCount);
-#else // OpenCL
-//    for (int i = 0; i < count; i++) {
-//        kernels->RemoveFactorsDynamicScaling(dScalingFactors[scalingIndices[i]], dScalingFactors[cumulativeScalingIndex],
-//                                             1, kPaddedPatternCount);
-//    }    
-#endif
     
 #ifdef BEAGLE_DEBUG_SYNCH    
     gpu->Synchronize();
@@ -1610,7 +1669,7 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::calculateRootLogLikelihoods(const int* bu
 
         *outSumLogLikelihood = 0.0;
         for (int i = 0; i < kSumSitesBlockCount; i++) {
-            if (!(hLogLikelihoodsCache[i] - hLogLikelihoodsCache[i] == 0.0))
+            if (hLogLikelihoodsCache[i] != hLogLikelihoodsCache[i])
                 returnCode = BEAGLE_ERROR_FLOATING_POINT;
             
             *outSumLogLikelihood += hLogLikelihoodsCache[i];
@@ -1667,7 +1726,7 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::calculateRootLogLikelihoods(const int* bu
             
             *outSumLogLikelihood = 0.0;
             for (int i = 0; i < kSumSitesBlockCount; i++) {
-                if (!(hLogLikelihoodsCache[i] - hLogLikelihoodsCache[i] == 0.0))
+                if (hLogLikelihoodsCache[i] != hLogLikelihoodsCache[i])
                     returnCode = BEAGLE_ERROR_FLOATING_POINT;
                 
                 *outSumLogLikelihood += hLogLikelihoodsCache[i];
@@ -1782,7 +1841,7 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::calculateEdgeLogLikelihoods(const int* pa
             
             *outSumLogLikelihood = 0.0;
             for (int i = 0; i < kSumSitesBlockCount; i++) {
-                if (!(hLogLikelihoodsCache[i] - hLogLikelihoodsCache[i] == 0.0))
+                if (hLogLikelihoodsCache[i] != hLogLikelihoodsCache[i])
                     returnCode = BEAGLE_ERROR_FLOATING_POINT;
                 
                 *outSumLogLikelihood += hLogLikelihoodsCache[i];
@@ -1831,7 +1890,7 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::calculateEdgeLogLikelihoods(const int* pa
             
             *outSumLogLikelihood = 0.0;
             for (int i = 0; i < kSumSitesBlockCount; i++) {
-                if (!(hLogLikelihoodsCache[i] - hLogLikelihoodsCache[i] == 0.0))
+                if (hLogLikelihoodsCache[i] != hLogLikelihoodsCache[i])
                     returnCode = BEAGLE_ERROR_FLOATING_POINT;
                 
                 *outSumLogLikelihood += hLogLikelihoodsCache[i];
@@ -1888,7 +1947,7 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::calculateEdgeLogLikelihoods(const int* pa
             
             *outSumLogLikelihood = 0.0;
             for (int i = 0; i < kSumSitesBlockCount; i++) {
-                if (!(hLogLikelihoodsCache[i] - hLogLikelihoodsCache[i] == 0.0))
+                if (hLogLikelihoodsCache[i] != hLogLikelihoodsCache[i])
                     returnCode = BEAGLE_ERROR_FLOATING_POINT;
                 
                 *outSumLogLikelihood += hLogLikelihoodsCache[i];
@@ -1974,7 +2033,7 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::calculateEdgeLogLikelihoods(const int* pa
                 
                 *outSumLogLikelihood = 0.0;
                 for (int i = 0; i < kSumSitesBlockCount; i++) {
-                    if (!(hLogLikelihoodsCache[i] - hLogLikelihoodsCache[i] == 0.0))
+                    if (hLogLikelihoodsCache[i] != hLogLikelihoodsCache[i])
                         returnCode = BEAGLE_ERROR_FLOATING_POINT;
                     
                     *outSumLogLikelihood += hLogLikelihoodsCache[i];
@@ -1995,20 +2054,16 @@ int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::calculateEdgeLogLikelihoods(const int* pa
     return returnCode;
 }
 
-template<>
-int BeagleGPUImpl<float>::getSiteLogLikelihoods(double* outLogLikelihoods) {
+BEAGLE_GPU_TEMPLATE
+int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::getSiteLogLikelihoods(double* outLogLikelihoods) {
 
 #ifdef BEAGLE_DEBUG_FLOW
     fprintf(stderr, "\tEntering BeagleGPUImpl::getSiteLogLikelihoods\n");
 #endif
 
-//#ifdef DOUBLE_PRECISION
-//    gpu->MemcpyDeviceToHost(outLogLikelihoods, dIntegrationTmp, sizeof(Real) * kPatternCount);
-//#else
-    gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dIntegrationTmp, sizeof(float) * kPatternCount);
-//    MEMCNV(outLogLikelihoods, hLogLikelihoodsCache, kPatternCount, double);
+// TODO: copy directly to outLogLikelihoods when GPU is running in double precision
+    gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dIntegrationTmp, sizeof(Real) * kPatternCount);
     beagleMemCpy(outLogLikelihoods, hLogLikelihoodsCache, kPatternCount);
-//#endif
 
 #ifdef BEAGLE_DEBUG_FLOW
     fprintf(stderr, "\tLeaving  BeagleGPUImpl::getSiteLogLikelihoods\n");
@@ -2017,90 +2072,27 @@ int BeagleGPUImpl<float>::getSiteLogLikelihoods(double* outLogLikelihoods) {
     return BEAGLE_SUCCESS;
 }
 
-template<>
-int BeagleGPUImpl<double>::getSiteLogLikelihoods(double* outLogLikelihoods) {
-    
-#ifdef BEAGLE_DEBUG_FLOW
-    fprintf(stderr, "\tEntering BeagleGPUImpl::getSiteLogLikelihoods\n");
-#endif
-    
-//#ifdef DOUBLE_PRECISION
-    gpu->MemcpyDeviceToHost(outLogLikelihoods, dIntegrationTmp, sizeof(double) * kPatternCount);
-//#else
-//    gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dIntegrationTmp, sizeof(Real) * kPatternCount);
-//    MEMCNV(outLogLikelihoods, hLogLikelihoodsCache, kPatternCount, double);
-//#endif
-    
-#ifdef BEAGLE_DEBUG_FLOW
-    fprintf(stderr, "\tLeaving  BeagleGPUImpl::getSiteLogLikelihoods\n");
-#endif    
-    
-    return BEAGLE_SUCCESS;
-}
 
-template<>
-int BeagleGPUImpl<double>::getSiteDerivatives(double* outFirstDerivatives,
+BEAGLE_GPU_TEMPLATE
+int BeagleGPUImpl<BEAGLE_GPU_GENERIC>::getSiteDerivatives(double* outFirstDerivatives,
                                       double* outSecondDerivatives) {
 
 #ifdef BEAGLE_DEBUG_FLOW
     fprintf(stderr, "\tEntering BeagleGPUImpl::getSiteDerivatives\n");
 #endif
 
-//#ifdef DOUBLE_PRECISION
-    gpu->MemcpyDeviceToHost(outFirstDerivatives, dOutFirstDeriv, sizeof(double) * kPatternCount);
-//#else
-//    gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dOutFirstDeriv, sizeof(Real) * kPatternCount);
-//    MEMCNV(outFirstDerivatives, hLogLikelihoodsCache, kPatternCount, double);
-//#endif
-
-    if (outSecondDerivatives != NULL) {
-//#ifdef DOUBLE_PRECISION
-        gpu->MemcpyDeviceToHost(outSecondDerivatives, dOutSecondDeriv, sizeof(double) * kPatternCount);
-//#else
-//        gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dOutSecondDeriv, sizeof(Real) * kPatternCount);
-//        MEMCNV(outSecondDerivatives, hLogLikelihoodsCache, kPatternCount, double);
-//#endif
-    }
-
-
-#ifdef BEAGLE_DEBUG_FLOW
-    fprintf(stderr, "\tLeaving  BeagleGPUImpl::getSiteDerivatives\n");
-#endif
-
-    return BEAGLE_SUCCESS;
-}
-
-template<>
-int BeagleGPUImpl<float>::getSiteDerivatives(double* outFirstDerivatives,
-                                      double* outSecondDerivatives) {
-    
-#ifdef BEAGLE_DEBUG_FLOW
-    fprintf(stderr, "\tEntering BeagleGPUImpl::getSiteDerivatives\n");
-#endif
-    
-//#ifdef DOUBLE_PRECISION
-//    gpu->MemcpyDeviceToHost(outFirstDerivatives, dOutFirstDeriv, sizeof(Real) * kPatternCount);
-//#else
-    gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dOutFirstDeriv, sizeof(float) * kPatternCount);
+    gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dOutFirstDeriv, sizeof(Real) * kPatternCount);
     beagleMemCpy(outFirstDerivatives, hLogLikelihoodsCache, kPatternCount);
-//    MEMCNV(outFirstDerivatives, hLogLikelihoodsCache, kPatternCount, double);
-//#endif
 
     if (outSecondDerivatives != NULL) {
-//#ifdef DOUBLE_PRECISION
-//        gpu->MemcpyDeviceToHost(outSecondDerivatives, dOutSecondDeriv, sizeof(Real) * kPatternCount);
-//#else
-        gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dOutSecondDeriv, sizeof(float) * kPatternCount);
+        gpu->MemcpyDeviceToHost(hLogLikelihoodsCache, dOutSecondDeriv, sizeof(Real) * kPatternCount);
         beagleMemCpy(outSecondDerivatives, hLogLikelihoodsCache, kPatternCount);
-//        MEMCNV(outSecondDerivatives, hLogLikelihoodsCache, kPatternCount, double);
-//#endif
     }
     
-    
 #ifdef BEAGLE_DEBUG_FLOW
     fprintf(stderr, "\tLeaving  BeagleGPUImpl::getSiteDerivatives\n");
-#endif    
-    
+#endif
+
     return BEAGLE_SUCCESS;
 }
 
@@ -2118,6 +2110,7 @@ BeagleImpl*  BeagleGPUImplFactory<BEAGLE_GPU_GENERIC>::createImpl(int tipCount,
                                               int categoryCount,
                                               int scaleBufferCount,
                                               int resourceNumber,
+                                              int pluginResourceNumber,
                                               long preferenceFlags,
                                               long requirementFlags,
                                               int* errorCode) {
@@ -2126,7 +2119,7 @@ BeagleImpl*  BeagleGPUImplFactory<BEAGLE_GPU_GENERIC>::createImpl(int tipCount,
         *errorCode =
             impl->createInstance(tipCount, partialsBufferCount, compactBufferCount, stateCount,
                                  patternCount, eigenBufferCount, matrixBufferCount,
-                                 categoryCount,scaleBufferCount, resourceNumber, preferenceFlags, requirementFlags);      
+                                 categoryCount,scaleBufferCount, resourceNumber, pluginResourceNumber, preferenceFlags, requirementFlags);      
         if (*errorCode == BEAGLE_SUCCESS) {
             return impl;
         }
@@ -2144,15 +2137,27 @@ BeagleImpl*  BeagleGPUImplFactory<BEAGLE_GPU_GENERIC>::createImpl(int tipCount,
     return NULL;
 }
 
+#ifdef CUDA
 template<>
 const char* BeagleGPUImplFactory<double>::getName() {
-	return "GPU-DP";
+    return "GPU-DP-CUDA";
 }
 
 template<>
 const char* BeagleGPUImplFactory<float>::getName() {
-	return "GPU-SP";
+    return "GPU-SP-CUDA";
 }
+#elif defined(FW_OPENCL)
+template<>
+const char* BeagleGPUImplFactory<double>::getName() {
+    return "DP-OpenCL";
+
+}
+template<>
+const char* BeagleGPUImplFactory<float>::getName() {
+    return "SP-OpenCL";
+}
+#endif
 
 template<>
 void modifyFlagsForPrecision(long *flags, double r) {
@@ -2170,15 +2175,23 @@ const long BeagleGPUImplFactory<BEAGLE_GPU_GENERIC>::getFlags() {
           BEAGLE_FLAG_SCALING_MANUAL | BEAGLE_FLAG_SCALING_ALWAYS | BEAGLE_FLAG_SCALING_AUTO | BEAGLE_FLAG_SCALING_DYNAMIC |
           BEAGLE_FLAG_THREADING_NONE |
           BEAGLE_FLAG_VECTOR_NONE |
-          BEAGLE_FLAG_PROCESSOR_GPU |
           BEAGLE_FLAG_SCALERS_LOG | BEAGLE_FLAG_SCALERS_RAW |
           BEAGLE_FLAG_EIGEN_COMPLEX | BEAGLE_FLAG_EIGEN_REAL |
           BEAGLE_FLAG_INVEVEC_STANDARD | BEAGLE_FLAG_INVEVEC_TRANSPOSED;
 
+#ifdef CUDA
+    flags |= BEAGLE_FLAG_FRAMEWORK_CUDA |
+             BEAGLE_FLAG_PROCESSOR_GPU;
+#elif defined(FW_OPENCL)
+    flags |= BEAGLE_FLAG_FRAMEWORK_OPENCL |
+             BEAGLE_FLAG_PROCESSOR_CPU | BEAGLE_FLAG_PROCESSOR_GPU | BEAGLE_FLAG_PROCESSOR_OTHER;
+#endif
+
 	Real r = 0;
 	modifyFlagsForPrecision(&flags, r);
 	return flags;
 }
 
+} // end of device namespace
 } // end of gpu namespace
 } // end of beagle namespace
