@@ -11,6 +11,28 @@
 #include "libhmsbeagle/beagle.h"
 #include "libhmsbeagle/JNI/beagle_BeagleJNIWrapper.h"
 
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    getVersion
+ * Signature: ()Ljava/lang/String;
+ */
+JNIEXPORT jstring JNICALL Java_beagle_BeagleJNIWrapper_getVersion
+  (JNIEnv *env, jobject obj)
+{
+    return env->NewStringUTF(beagleGetVersion());
+}
+
+/*
+ * Class:     beagle_BeagleJNIWrapper
+ * Method:    getCitation
+ * Signature: ()Ljava/lang/String;
+ */
+JNIEXPORT jstring JNICALL Java_beagle_BeagleJNIWrapper_getCitation
+  (JNIEnv *env, jobject obj)
+{
+    return env->NewStringUTF(beagleGetCitation());
+}
+
 /*
  * Class:     beagle_BeagleJNIWrapper
  * Method:    getResourceList
