@@ -49,6 +49,11 @@
 
 #include "libhmsbeagle/plugin/Plugin.h"
 
+#define BEAGLE_VERSION  PACKAGE_VERSION
+#define BEAGLE_CITATION "Using BEAGLE library v" PACKAGE_VERSION " for accelerated, parallel likelihood evaluation\n\
+2009-2013, BEAGLE Working Group - http://beagle-lib.googlecode.com/\n\
+Citation: Ayres et al (2012) Systematic Biology 61: 170-173 | doi:10.1093/sysbio/syr100\n"
+
 typedef std::list< std::pair<int,int> > PairedList;
 typedef std::pair<int, std::pair<int, beagle::BeagleImplFactory*> >	RsrcImpl;
 typedef std::list<RsrcImpl> RsrcImplList;
@@ -78,6 +83,7 @@ bool compareRsrcImpl(const RsrcImpl &left, const RsrcImpl &right) {
 std::list<beagle::BeagleImplFactory*>* implFactory = NULL;
 
 BeagleResourceList* rsrcList = NULL;
+std::map<int, int> ResourceMap;
 
 int loaded = 0; // Indicates is the initial library constructors have been run
                 // This patches a bug with JVM under Linux that calls the finalizer twice
@@ -111,10 +117,20 @@ void beagleLoadPlugins(void) {
 		plugins->push_back(openclplug);
 	}catch(beagle::plugin::SharedLibraryException sle){}
 
+	try{
+		beagle::plugin::Plugin* openclalteraplug = pm.findPlugin("hmsbeagle-opencl-altera");
+		plugins->push_back(openclalteraplug);
+	}catch(beagle::plugin::SharedLibraryException sle){}
+
 	try{
 		beagle::plugin::Plugin* sseplug = pm.findPlugin("hmsbeagle-cpu-sse");
 		plugins->push_back(sseplug);
 	}catch(beagle::plugin::SharedLibraryException sle){}
+	
+	try{
+		beagle::plugin::Plugin* avxplug = pm.findPlugin("hmsbeagle-cpu-avx");
+		plugins->push_back(avxplug);
+	}catch(beagle::plugin::SharedLibraryException sle){}	
 
 	try{
 		beagle::plugin::Plugin* openmpplug = pm.findPlugin("hmsbeagle-cpu-openmp");
@@ -207,6 +223,14 @@ int beagleFinalize() {
     return BEAGLE_SUCCESS;
 }
 
+const char* beagleGetVersion() {
+    return BEAGLE_VERSION;
+}
+
+const char* beagleGetCitation() {
+    return BEAGLE_CITATION;
+}
+
 BeagleResourceList* beagleGetResourceList() {
 
 	// plugins must be loaded before resources
@@ -224,7 +248,7 @@ BeagleResourceList* beagleGetResourceList() {
         
         // allocate space for a complete list of resources
         rsrcList->list = (BeagleResource*) malloc(sizeof(BeagleResource) * rsrcList->length);
-        
+
         // copy in resource lists from each plugin
         int rI=0;
         for(plugin_iter = plugins->begin(); plugin_iter != plugins->end(); plugin_iter++ ){
@@ -243,8 +267,10 @@ BeagleResourceList* beagleGetResourceList() {
                     }
                 }
                 
-                if (!rsrcExists)
+                if (!rsrcExists) {
+                    ResourceMap.insert(std::pair<int, int>(rI, (rI - prev_rI)));
                     rsrcList->list[rI++] = *r_iter;
+                }
             }
         }
     }
@@ -351,7 +377,7 @@ int beagleCreateInstance(int tipCount,
 #endif
                 if ( ((requirementFlags & factoryFlags) >= requirementFlags) // Factory meets requirementFlags
                     && ((resourceRequiredFlags & factoryFlags) >= resourceRequiredFlags) // Factory meets resourceFlags
-                    && ((resourceSupportedFlags & factoryFlags) >= factoryFlags) // Resource meets factoryFlags
+                    && ((requirementFlags & resourceSupportedFlags) >= requirementFlags) // Resource meets requirementFlags
                     ) {
                     int implementationScore = scoreFlags(preferenceFlags,factoryFlags);
                     int totalScore = resourceScore + implementationScore;
@@ -398,6 +424,7 @@ int beagleCreateInstance(int tipCount,
                                                                 matrixBufferCount, categoryCount,
                                                                 scaleBufferCount,
                                                                 resource,
+                                                                ResourceMap[resource],
                                                                 preferenceFlags,
                                                                 requirementFlags,
                                                                 &errorCode);
