@@ -1,7 +1,7 @@
 /**
  * @file beagle.h
  *
- * Copyright 2009-2011 Phylogenetic Likelihood Working Group
+ * Copyright 2009-2013 Phylogenetic Likelihood Working Group
  *
  * This file is part of BEAGLE.
  *
@@ -26,7 +26,7 @@
  * 
  * The key to BEAGLE performance lies in delivering fine-scale
  * parallelization while minimizing data transfer and memory copy overhead.
- * To accomplish this, the library lacks the concept or data structure for
+ * To accomplish this, the library lacks the concept of data structure for
  * a tree, in spite of the intended use for phylogenetic analysis. Instead,
  * BEAGLE acts directly on flexibly indexed data storage (called buffers)
  * for observed character states and partial likelihoods. The client
@@ -142,17 +142,17 @@ enum BeagleReturnCodes {
 enum BeagleFlags {
     BEAGLE_FLAG_PRECISION_SINGLE    = 1 << 0,    /**< Single precision computation */
     BEAGLE_FLAG_PRECISION_DOUBLE    = 1 << 1,    /**< Double precision computation */
-
+    
     BEAGLE_FLAG_COMPUTATION_SYNCH   = 1 << 2,    /**< Synchronous computation (blocking) */
     BEAGLE_FLAG_COMPUTATION_ASYNCH  = 1 << 3,    /**< Asynchronous computation (non-blocking) */
     
     BEAGLE_FLAG_EIGEN_REAL          = 1 << 4,    /**< Real eigenvalue computation */
     BEAGLE_FLAG_EIGEN_COMPLEX       = 1 << 5,    /**< Complex eigenvalue computation */
-
+    
     BEAGLE_FLAG_SCALING_MANUAL      = 1 << 6,    /**< Manual scaling */
     BEAGLE_FLAG_SCALING_AUTO        = 1 << 7,    /**< Auto-scaling on */
     BEAGLE_FLAG_SCALING_ALWAYS      = 1 << 8,    /**< Scale at every updatePartials */
-    BEAGLE_FLAG_SCALING_DYNAMIC     = 1 << 19,   /**< Manual scaling with dynamic checking  */
+    BEAGLE_FLAG_SCALING_DYNAMIC     = 1 << 25,   /**< Manual scaling with dynamic checking  */
     
     BEAGLE_FLAG_SCALERS_RAW         = 1 << 9,    /**< Save raw scalers */
     BEAGLE_FLAG_SCALERS_LOG         = 1 << 10,   /**< Save log scalers */
@@ -161,6 +161,7 @@ enum BeagleFlags {
     BEAGLE_FLAG_INVEVEC_TRANSPOSED  = 1 << 21,   /**< Inverse eigen vectors passed to BEAGLE have been transposed */
     
     BEAGLE_FLAG_VECTOR_SSE          = 1 << 11,   /**< SSE computation */
+    BEAGLE_FLAG_VECTOR_AVX          = 1 << 24,   /**< AVX computation */
     BEAGLE_FLAG_VECTOR_NONE         = 1 << 12,   /**< No vector computation */
     
     BEAGLE_FLAG_THREADING_OPENMP    = 1 << 13,   /**< OpenMP threading */
@@ -170,9 +171,12 @@ enum BeagleFlags {
     BEAGLE_FLAG_PROCESSOR_GPU       = 1 << 16,   /**< Use GPU as main processor */
     BEAGLE_FLAG_PROCESSOR_FPGA      = 1 << 17,   /**< Use FPGA as main processor */
     BEAGLE_FLAG_PROCESSOR_CELL      = 1 << 18,   /**< Use Cell as main processor */
+    BEAGLE_FLAG_PROCESSOR_PHI       = 1 << 19,   /**< Use Intel Phi as main processor */
+    BEAGLE_FLAG_PROCESSOR_OTHER     = 1 << 26,   /**< Use other type of processor */
     
     BEAGLE_FLAG_FRAMEWORK_CUDA      = 1 << 22,   /**< Use CUDA implementation with GPU resources */
-    BEAGLE_FLAG_FRAMEWORK_OPENCL    = 1 << 23    /**< Use OpenCL implementation with GPU resources */
+    BEAGLE_FLAG_FRAMEWORK_OPENCL    = 1 << 23,    /**< Use OpenCL implementation with GPU resources */
+    BEAGLE_FLAG_FRAMEWORK_CPU       = 1 << 27   /**< Use CPU implementation */
 };
 
 /**
@@ -225,7 +229,26 @@ typedef struct {
 #ifdef __cplusplus
 extern "C" {
 #endif
-    
+
+/**
+ * @brief Get version
+ *
+ * This function returns a pointer to a string with the library version number.
+ *
+ * @return A string with the version number
+ */
+BEAGLE_DLLEXPORT const char* beagleGetVersion(void);
+
+/**
+ * @brief Get citation
+ *
+ * This function returns a pointer to a string describing the version of the
+ * library and how to cite it.
+ *
+ * @return A string describing the version of the library and how to cite it
+ */
+BEAGLE_DLLEXPORT const char* beagleGetCitation(void);
+
 /**
  * @brief Get list of hardware resources
  *
